import inspect
import math
import random
import re
import string

from boltons.typeutils import make_sentinel

NONE_TYPE = type(None)
MISSING = make_sentinel(var_name='MISSING')


def log10_ceil(num):
    return math.ceil(math.log10(num or 1)) or 1


def wraps_with_name(name):
    def wrapper(func):
        func.__name__ = func.__qualname__ = name
        return func

    return wrapper


def _internal_to_kebab_case(name: str, *, replacer='-'):
    return re.sub(
        r'(?P<upper>[A-Z])', lambda m: replacer + m.group('upper').lower(), re.sub('[_ ]', replacer, name.strip('_ '))
    ).strip(replacer)


def to_kebab_case(name: str, *, replacer='-'):
    """
    Converts name of tag to html-compatible or name of property to css-compatible

    >>> class __beepy__(Tag): ...  # <beepy></beepy>
    >>> class myTagName(Tag): ...  # <my-tag-name/>
    >>> Style(font_size='20px')  # font-size: 20px
    >>> Style(backgroundColor='red')  # background-color: red
    """
    if hasattr(name, '__html__'):
        return name.__html__()
    return _internal_to_kebab_case(name, replacer=replacer)


def escape_html(s, *, quote=False, whitespace=False):
    """Replace special characters "&", "<" and ">" to HTML-safe sequences.
    If the optional flag quote is True (default), the quotation mark characters (" and ') are also translated.
    If the optional flag whitespace is True, new line (\n) and tab (\t) characters are also translated.
    """
    #       Must be done first!
    s = s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')
    if quote:
        s = s.replace('"', '&quot;').replace("'", '&#39;')
    if whitespace:
        s = s.replace('\n', '<br>').replace('\t', '&emsp;').replace('  ', ' &nbsp;')
    return s


_w = string.ascii_letters + string.digits + '_'


def get_random_name(length=6):
    return ''.join(random.choice(_w) for _ in range(length))


def safe_issubclass(type_or_Any, class_or_tuple_to_check):
    try:
        return issubclass(type_or_Any, class_or_tuple_to_check)
    except TypeError:
        return False


class AnyOfType:
    __slots__ = ('type',)

    def __init__(self, type):
        self.type = type

    def __eq__(self, other):
        return isinstance(other, self.type)

    def __repr__(self):
        return f'AnyOfType<{self.type}>'

    def __hash__(self):
        return hash(self.type)


class Locker:
    __slots__ = ('name', 'locked')

    def __init__(self, name='Lock'):
        self.name = name
        self.locked = False

    def __enter__(self):
        self.locked = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.locked = False

    def __bool__(self):
        return self.locked

    def __str__(self):
        return f'Locker<{self.name}>({self.locked})'


def call_handler_with_optional_arguments(handler, obj, optional_args, *args):
    parameters = inspect.signature(handler).parameters

    for name, arg in reversed(optional_args.items()):
        if name in parameters:
            args = (arg, *args)
    return handler(obj, *args)


def nested_copy(dct):
    # We shouldn't use deepcopy, we want just copy nested dicts, not objects inside
    return {key: value.copy() for key, value in dct.items()}


__all__ = [
    'NONE_TYPE',
    'log10_ceil',
    'escape_html',
    'wraps_with_name',
    'get_random_name',
    'to_kebab_case',
    'safe_issubclass',
    'AnyOfType',
    'Locker',
    'call_handler_with_optional_arguments',
    'nested_copy',
]



from __future__ import annotations

from collections import defaultdict
from functools import partial, wraps
from types import MethodType
from typing import TYPE_CHECKING, Any

import beepy
from beepy.attrs import set_html_attribute, state
from beepy.context import Context, _MetaContext
from beepy.listeners import on
from beepy.types import AttrType, Renderer, WebBase
from beepy.utils import IN_BROWSER, js
from beepy.utils.common import NONE_TYPE, nested_copy
from beepy.utils.dev import _debugger

if TYPE_CHECKING:
    from collections.abc import Callable

    from beepy.children import ComponentRef
    from beepy.framework import Tag

_current__lifecycle_method: dict[str, dict[int, Component]] = {}


def _lifecycle_method(*, hash_function=hash):
    def _wrapper(fn):
        name = fn.__name__
        attr_name = f'_wrapper_{name}_calling'
        _cache = _current__lifecycle_method[attr_name] = {}

        @wraps(fn)
        def lifecycle_method(original_func):
            @wraps(original_func)
            def original_method_wrapper(self, *args, **kwargs):
                # prevent calling super() calls extra code twice
                _hash = hash_function(self)
                not_in_super_call = _hash not in _cache

                if not_in_super_call:
                    _cache[_hash] = self
                    result = fn(self, args, kwargs, _original_func=original_func)
                    del _cache[_hash]
                else:
                    result = original_func(self, *args, **kwargs)

                return result

            return original_method_wrapper

        return lifecycle_method

    return _wrapper


_COMPONENT_INITIALIZED = False


class _MetaComponent(_MetaContext):
    def __new__(mcs, _name: str, bases: tuple, namespace: dict, **kwargs):
        initialized = _COMPONENT_INITIALIZED  # As base classes is also declared here, we must be sure base class exists

        static_onchange_handlers = []

        if initialized:
            for _attribute_name, child in tuple(mcs._clean_namespace(namespace)):
                if not (callable(child) and hasattr(child, '_attrs_static_')):
                    continue

                _states_with_static_handler = defaultdict(list)
                for trigger, _states in child._attrs_static_.items():
                    for _state in _states:
                        _states_with_static_handler[trigger].append(_state)
                        _state.handlers[trigger].remove(child)
                static_onchange_handlers.append((child, _states_with_static_handler))

        cls: type[Component] | type = super().__new__(mcs, _name, bases, namespace, **kwargs)

        if initialized:
            cls._static_listeners = defaultdict(list, **nested_copy(cls._static_listeners))
            cls._static_onchange_handlers = cls._static_onchange_handlers.copy() + static_onchange_handlers
        else:
            cls._static_listeners = defaultdict(list)
            cls._static_onchange_handlers = []

        if hasattr(cls, '__extra_attributes__'):
            cls.__extra_attributes__ = {
                key: value.as_child(None) if isinstance(value, Component) else value
                for key, value in cls.__extra_attributes__.items()
            }

        if '__mount__' in namespace:
            cls.__mount__ = mcs.__mount(cls.__mount__)

        if '__render__' in namespace:
            cls.__render__ = mcs.__render(cls.__render__)

        if '__init__' in namespace:
            cls.__init__ = mcs.__init(cls.__init__)

        if '__unmount__' in namespace:
            cls.__unmount__ = mcs.__unmount(cls.__unmount__)

        return cls

    @_lifecycle_method(hash_function=id)
    def __init(self: Component, args, kwargs, _original_func):
        return self.__class__.__META_init__(self, args, kwargs, _original_func)

    @_lifecycle_method()
    def __mount(self: Component, args, kwargs, _original_func):
        return self.__class__.__META_mount__(self, args, kwargs, _original_func)

    @_lifecycle_method()
    def __unmount(self: Component, args, kwargs, _original_func):
        return self.__class__.__META_unmount__(self, args, kwargs, _original_func)

    @_lifecycle_method()
    def __render(self: Component, args, kwargs, _original_func):
        return self.__class__.__META_render__(self, args, kwargs, _original_func)

    @classmethod
    def __META_init__(cls, self: Component, args, kwargs, _original_func):
        self._parent_ = None

        self._dependents = []
        self._listeners = defaultdict(list, **nested_copy(self._static_listeners))
        self._event_listeners = defaultdict(list)
        self._handlers = defaultdict(list)

        self._ref = None

        _original_func(self, *args, **kwargs)

    @classmethod
    def __META_mount__(cls, self: Component, args, kwargs, _original_func, *, _mount_attrs=True):
        result = _original_func(self, *args, **kwargs) if _original_func else None

        if _mount_attrs:
            self._mount_attrs()

        for name, attribute in self.__states__.items():
            if callable(attribute) and not isinstance(attribute, MethodType):
                setattr(self, name, MethodType(attribute, self.parent))

        self._post_mount_attrs()

        for event, listeners in self._listeners.items():
            for listener in listeners:
                self._event_listeners[event].append(listener._make_listener(event, self))

        for onchange_handler, _states_with_static_handler in self._static_onchange_handlers:
            for trigger, _states in _states_with_static_handler.items():
                for _state in _states:
                    # TODO: can we save order of triggers' call?
                    _state.handlers[trigger].append(MethodType(onchange_handler, self))

        self.mount()

        return result

    @classmethod
    def __META_unmount__(cls, self: Component, args, kwargs, _original_func, *, _post_mount=True):
        if not _post_mount:
            self.unmount()

        result = _original_func(self, *args, **kwargs) if _original_func else None

        if IN_BROWSER:
            for event, event_listeners in self._event_listeners.items():
                for event_listener in event_listeners:
                    on._remove_listener(event, self, event_listener)

        if _post_mount:
            self.unmount()

        return result

    @classmethod
    def __META_render__(cls, self: Component, args, kwargs, _original_func):
        # TODO: maybe function 'render' could return some content, appended to args?
        self.render()

        attrs: dict[str, AttrType] = args[0] if len(args) > 0 else {}

        for name, value in {**self.__attrs__, **attrs}.items():
            # TODO: optimize this - set only changed attributes

            type = _attr.type if (_attr := self.attrs.get(name)) else NONE_TYPE
            set_html_attribute(self.mount_element, name, value, type=type)

        self.post_render()

        _original_func(self, *args, **kwargs)


class Component(WebBase, Context, metaclass=_MetaComponent, _root=True):
    __slots__ = ('_parent_', '_event_listeners', '_dependents', '_listeners', '_handlers', '_ref')

    parent: Tag | None
    mount_element: js.HTMLElement | None

    _ref: ComponentRef | None
    _force_ref: bool

    _event_listeners: defaultdict[str, list[Callable[[js.Event], None]]]
    _static_listeners: defaultdict[str, list[on]]
    _dependents: list[Renderer]
    _listeners: defaultdict[str, list[on]]
    _handlers: defaultdict[str, list[Callable[[Tag, js.Event, str, Any], None]]]
    _static_onchange_handlers: list[tuple[Callable[[Tag, Any], Any], dict[str, list[state]]]]

    @property
    def parent_defined(self):
        return self._parent_ is not None

    @property
    def parent(self):
        if self._parent_ is None:
            _debugger("ValueError: Trying to get .parent, but it's undefined")
        return self._parent_

    @parent.setter
    def parent(self, v):
        self._parent_ = v

    def as_child(self, parent: Tag | None, *, exists_ok=False):
        if self._ref:
            if exists_ok:
                self.__set_ref__(parent, self._ref)
                return self._ref
            else:
                raise TypeError(f'Component {self._context_name_} already is child')
        ref = beepy.children.ComponentRef(self)
        self.__set_ref__(parent, ref)
        return ref

    def __set_ref__(self, parent: Tag | None, ref: ComponentRef):
        self._ref = ref

    def clone(self, parent=None) -> Component:
        clone = super().clone(parent=parent)
        clone._listeners = defaultdict(list, **nested_copy(self._listeners))
        clone._handlers = defaultdict(list, **nested_copy(self._handlers))
        return clone

    def __init__(self, *args, **kwargs: AttrType):
        # DO NOT DELETE; This method must be wrapped by _MetaTag.__init
        super().__init__(*args, **kwargs)

    def __mount__(self, element, parent: Tag, index=None):
        self.parent = parent
        self.pre_mount()

        args, kwargs = self.args_kwargs
        kwargs = self._attrs_defaults | kwargs
        self.init(*args, **kwargs)

    def _mount_attrs(self):
        for attribute in self.attrs.values():
            attribute.__mount_cmpt__(self)

    def _post_mount_attrs(self):
        for attribute in self.attrs.values():
            attribute.__post_mount_cmpt__(self)

    def pre_mount(self):
        """empty method for easy override with code for run before mount"""

    def mount(self):
        """empty method for easy override with code for run after mount"""

    def unmount(self):
        """empty method for easy override with code for run before unmount"""

    def post_unmount(self):
        """empty method for easy override with code for run after unmount"""

    def render(self):
        """empty method for easy override with code for run before render"""

    def post_render(self):
        """empty method for easy override with code for run after render"""

    def on(self, method: Callable | str):
        if isinstance(method, str) and method.startswith(':'):  # TODO: maybe it could be useful in `class on()`?
            action = method[1:]

            def wrapper(handler, action_name=None):
                if action_name is None:
                    action_name = handler.__name__

                self._handlers[action_name].append(handler)
                return handler

            if action:
                return partial(wrapper, action_name=action)

            return wrapper

        def wrapper(callback: Callable):
            event_listener = on(method)(callback, child=self)
            event_name = event_listener.name or callback.__name__
            event_listener.__set_name__(self, event_name)
            self._listeners[event_name] = [*self._listeners[event_name].copy(), event_listener]
            return callback

        if not isinstance(method, str):
            return wrapper(method)

        return wrapper


_COMPONENT_INITIALIZED = True


class Directive(Component, _root=True):
    _force_ref: bool = True

    element = state()

    def __unmount__(self, element, parent, *, _unsafe=False):
        # DO NOT DELETE; This method must be wrapped by _MetaTag.__unmount
        pass

    def __render__(self):
        # DO NOT DELETE; This method must be wrapped by _MetaTag.__render
        pass

    @property
    def el(self):
        if self.element:
            return getattr(self.parent, self.element)
        return self.parent

    @property
    def mount_element(self):
        return self.parent.mount_element


__all__ = ['_MetaComponent', 'Component', 'Directive']



from collections.abc import Callable

from beepy import Tag
from beepy.attrs import attr, state
from beepy.listeners import on
from beepy.style import Style
from beepy.tags import Body, button, div


class Modal(Tag, name='modal', content_tag='h2', children_tag='modal-content'):
    visible = attr(default=False)
    on_close = state(type=Callable[[], None])  # TODO: add example

    default_style = Style(
        styles={
            'position': 'absolute',
            'inset': 0,
            'width': '100%',
            'height': '100%',
            'display': 'flex',
            ('justify-content', 'justify-items', 'align-items'): 'center',
            'background-color': 'rgba(255, 255, 255, 0.2)',
            'opacity': 0,
            'visibility': 'hidden',
            'transition': 'opacity 0.2s, visibility 0.2s',
            '&[visible]': {
                'opacity': 1,
                'visibility': 'visible',
            },
            'modal-content': {
                'position': 'absolute',
                'padding': '5%',
                ('width', 'height'): 'fit-content',
                'display': 'flex',
                'flex-direction': 'column',
                ('justify-content', 'justify-items', 'align-items'): 'center',
                'background-color': '#333',
                'box-shadow': '0 0 1.5rem rgb(255 255 255 / 33%)',  # TODO: create beepy.style.rgb function
            },
        }
    )

    children = [
        button_close := button('Close'),
    ]

    def show(self):
        self.visible = True

    @on('keyup.esc')
    @button_close.on('click')
    def close(self):
        self.visible = False
        if self.on_close:
            self.on_close()

    def __mount__(self, element, parent, index=None):
        super().__mount__(Body.mount_element, Body)

    def __unmount__(self, element, parent, *, _unsafe=False):
        return super().__unmount__(element, parent, _unsafe=True)


class ModalHandler(div, content_tag=div()):
    modal: Modal

    default_style = Style()

    children = [
        button_show := button('Show Modal'),
    ]

    @button_show.on('click')
    def show_modal(self):
        self.modal.show()



from __future__ import annotations

from typing import Literal, TypeVar

from beepy.attrs import state
from beepy.children import Children
from beepy.listeners import on
from beepy.modules.actions import Action
from beepy.style import Style
from beepy.tags import table, tbody, td, th, thead, tr
from beepy.utils.asyncio import ensure_sync_many
from beepy.utils.common import call_handler_with_optional_arguments

T = TypeVar('T')


class TableCellAction(Action, _root=True):
    components: dict[str, type[TableCellAction]] = {}

    parent: TD

    @on
    def click(self, event):
        tr: TR = self.parent.parent
        table: Table = tr.parent.parent
        row = table._map_data(tr.raw_data)
        index = table.find_row_index(id=row['id'])

        ensure_sync_many(
            [
                call_handler_with_optional_arguments(
                    handler, table.parent, {'event': event, 'action': self.action_name, 'index': index}, row
                )
                for handler in table._handlers[self.action_name]
            ],
            lambda *_: table.parent.__render__(),
        )


class ActionEdit(TableCellAction):
    action_name = 'edit'
    _content = 'Edit'


class ActionDelete(TableCellAction):
    action_name = 'delete'
    _content = 'Delete'


class TH(th):
    parent: TR


class TD(td):
    parent: TR


class TR(tr):
    data = state(type=list[dict[str, str]])

    parent: TableHead | TableBody

    children = [
        _data := Children(),  # TODO: add auto-reload data, instead of `sync` functions
    ]

    @property
    def raw_data(self):
        return [line['value'] for line in self.data if isinstance(line, dict)]

    @property
    def view_data(self):
        return [(line['view'] if isinstance(line, dict) else line) for line in self.data]

    @data.on('mount', 'change')
    def sync(self):
        self._data[:] = [(TD(str(line['view'])) if isinstance(line, dict) else TD(*line)) for line in self.data]


class TableHead(thead, children_tag=TR(), force_ref=True):
    columns = state(type=list[dict[Literal['id', 'label', 'view'], str]])
    # TODO: maybe use 'name' instead of 'id'
    #       add 'hidden' option (availability to hide some columns)

    parent: Table

    children = [
        _columns := Children(),
    ]

    @columns.on('mount', 'change')
    def sync(self):
        if not self.parent_defined:
            return

        self._columns[:] = [TH(column['label']) for column in self.columns]
        if self.parent.actions:
            self._columns.append(TH('Actions'))


class TableBody(tbody, force_ref=True):
    rows = state(type=list[list[str]])

    parent: Table

    children = [
        _rows := Children(),
    ]

    @rows.on('mount', 'change')
    def sync(self):
        if not self.parent_defined:
            return

        self._rows[:] = [
            TR(
                data=[
                    *(
                        {'value': cell, 'view': col['view'](cell).__view_value__() if 'view' in col else cell}
                        for col, cell in self.parent._zip_column_row(row)
                    ),
                    [TableCellAction.components[action]() for action in self.parent.actions],
                ]
            )
            for row in self.rows
        ]

    def delete_row(self, index):
        self.rows.pop(index)
        self._rows.pop(index)


class Table(table):
    actions = ('edit', 'delete')

    head: TableHead
    body: TableBody

    default_style = Style(
        styles={
            '&': {
                'border-spacing': 0,
                'border-collapse': 'collapse',
            },
            '& >': {
                '&, th, td': {
                    'border': '1px solid #333',
                },
                'th, td': {
                    'padding': '8px',
                },
            },
        }
    )

    def _zip_column_row(self, row):
        if isinstance(row, dict):
            for col in self.head.columns:
                yield col, row[col['id']]
        else:
            for index, col in enumerate(self.head.columns):
                yield col, row[index]

    def _map_data(self, row):
        if isinstance(row, dict):
            return row
        return dict(zip([col['id'] for col in self.head.columns], row, strict=True))

    @property
    def data(self):
        return [self._map_data(row) for row in self.body.rows]

    def find_row_index(self, *, raise_key_error=True, **filters):
        if '_index' in filters:
            return filters['_index']

        row_index = None

        for index, row in enumerate(self.data):
            for key, value in filters.items():
                if row[key] == value:
                    row_index = index
                    break
            else:
                continue
            break

        if raise_key_error and row_index is None:
            raise KeyError(str(filters))

        return row_index

    def find_row(self, **filters):
        return self.body.rows[self.find_row_index(**filters)]

    def delete_row(self, **filters):
        self.body.delete_row(self.find_row_index(**filters))



import sys
from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location
from pathlib import Path

from pyodide.ffi import JsException

from beepy.utils.dev import _debugger
from beepy.utils.internal import __CONFIG__, BEEPY_ROOT_PACKAGE
from beepy.utils.js_py import IN_BROWSER, js

requirements = __CONFIG__['requirements']
MODULES_NOT_EXISTING_ON_SERVER = [
    BEEPY_ROOT_PACKAGE,
    *(requirements() if callable(requirements) else requirements),
    '_hashlib',  # TODO: FIX THIS...
    '_strptime',
    'unicodedata',
    'pprint',
    'numpy',
    'matplotlib',
    'numbers',
    'pickle5',
    'pickle',
    '_compat_pickle',
    '_pickle',
    'org',
    'ctypes',
    '_ctypes',
    'backports_abc',
    'secrets',
    'hmac',
    'gzip',
    'shlex',
    'defusedxml',
    'cffi',
    'uuid',
    '_uuid',
    'cycler',
    'six',
    'six.moves',
    'six.moves.winreg',
    'decimal',
    '_decimal',
    'http',
    'ssl',
]
# some modules must be ignored to prevent load it from local server, when importing modules like micropip or datetime


class ServerFinder(MetaPathFinder):
    def find_spec(self, fullname, path, target=None):  # noqa: ARG002 - override of MetaPathFinder
        if path and any(p.startswith('/lib') for p in path):
            return

        if Path(f'/lib/python3.11/site-packages/{fullname}').exists() or fullname in MODULES_NOT_EXISTING_ON_SERVER:
            return

        current_path = js.beepy.files._lastLoadedFile

        try:
            js.beepy.loadModule(fullname)
        except JsException as err:
            js.beepy.files._lastLoadedFile = current_path
            _debugger(err)
            MODULES_NOT_EXISTING_ON_SERVER.append(fullname)
            return

        return spec_from_file_location(fullname)


if IN_BROWSER:
    sys.meta_path.insert(0, ServerFinder())
    sys.path.append(BEEPY_ROOT_PACKAGE)



from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import SupportsIndex

from beepy.utils.common import Locker


class Trackable(ABC):
    __slots__ = ()

    onchange_triggers: list[Callable]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.onchange_triggers = []
        self._disable_onchange = Locker('Disable onchange()')

    @abstractmethod
    def onchange_notify(self):
        pass

    def onchange(self, handler=None):
        if handler is None:
            if not self._disable_onchange:
                self.onchange_notify()
        else:
            if handler in self.onchange_triggers:
                raise AttributeError('This @onchange trigger is already set')
            self.onchange_triggers.append(handler)
            return handler

    @abstractmethod
    def _notify_add_one(self, key: int, added):
        pass

    @abstractmethod
    def _notify_remove_one(self, key: int, to_remove):
        pass

    @abstractmethod
    def _notify_add(self, key: SupportsIndex | slice, added):
        pass

    @abstractmethod
    def _notify_remove(self, key: SupportsIndex | slice, to_remove):
        pass

    @abstractmethod
    def _notify_post_remove(self):
        pass


class TrackableList(Trackable, list):
    __slots__ = ('onchange_triggers', '_disable_onchange')

    def _notify_add(self, key: SupportsIndex | slice, added: Iterable):
        length = len(self)
        if isinstance(key, slice):
            for index, value in zip(range(key.start or 0, key.stop or length, key.step or 1), added, strict=True):
                if index < 0:
                    index += length  # noqa: PLW2901
                self._notify_add_one(index, value)
        else:
            index = key.__index__()
            if index < 0:
                index += length
            self._notify_add_one(index, added[0])
        if added:
            self.onchange()

    def _notify_remove(self, key: SupportsIndex | slice, to_remove: tuple | list):
        if isinstance(key, slice):
            for index, value in reversed(
                tuple(zip(range(key.start or 0, key.stop or len(self), key.step or 1), to_remove, strict=True))
            ):
                if index < 0:
                    index += len(self)  # noqa: PLW2901
                self._notify_remove_one(index, value)
        else:
            index = key.__index__()
            if index < 0:
                index += len(self)
            self._notify_remove_one(index, to_remove[0])

    def append(self, __object):
        super().append(__object)

        if not self._disable_onchange:
            self._notify_add(-1, (self[-1],))

    def clear(self):
        length = len(self)
        if not self._disable_onchange:
            self._notify_remove(slice(0, length), self)
        super().clear()
        if not self._disable_onchange and length:
            self._notify_post_remove()
            self.onchange()

    def extend(self, __iterable):
        if self._disable_onchange:
            super().extend(__iterable)
            return

        length = len(self)
        super().extend(__iterable)
        self._notify_add(slice(length, len(self)), self[length : len(self)])

    def insert(self, __index, __object):
        if self._disable_onchange:
            super().insert(__index, __object)
            return

        index = __index.__index__()
        super().insert(index, __object)
        self._notify_add(index, (self[index],))

    def pop(self, __index=None):
        if self._disable_onchange:
            return super().pop(__index)

        index = len(self) - 1 if __index is None else __index.__index__()

        self._notify_remove(index, (self[index],))
        result = super().pop(__index)
        self._notify_post_remove()
        self.onchange()
        return result

    def remove(self, __value):
        if self._disable_onchange:
            super().remove(__value)
            return

        self._notify_remove(self.index(__value), (__value,))
        super().remove(__value)
        self._notify_post_remove()
        self.onchange()

    def copy(self):
        result = type(self)(super().copy())
        result.onchange_triggers = self.onchange_triggers.copy()
        return result

    def reverse(self):
        if self._disable_onchange:
            super().reverse()
            return

        length = len(self)
        self._notify_remove(slice(0, length), self)

        super().reverse()

        if length:
            self._notify_post_remove()

        self._notify_add(slice(0, len(self)), self)

    def sort(self, *, key=..., reverse=...):
        if self._disable_onchange:
            super().sort(key=key, reverse=reverse)
            return

        # TODO: rewrite this after `key=` implemented
        length = len(self)
        self._notify_remove(slice(0, length), self)

        super().sort(key=key, reverse=reverse)

        if length:
            self._notify_post_remove()

        self._notify_add(slice(0, len(self)), self)

    def __delitem__(self, key):
        if self._disable_onchange:
            super().__delitem__(key)
            return

        to_remove = self[key]
        if not isinstance(to_remove, list):
            to_remove = (to_remove,)

        self._notify_remove(key, to_remove)
        super().__delitem__(key)
        if to_remove:
            self._notify_post_remove()
            self.onchange()

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __imul__(self, n):
        if self._disable_onchange:
            return super().__imul__(n)

        length = len(self)
        n = n.__index__()
        if n <= 0:
            self._notify_remove(slice(0, length), self)
        super().__imul__(n)
        if n <= 0:
            if length:
                self._notify_post_remove()
        else:
            self._notify_add(slice(length, len(self)), self[length : len(self)])
        return self

    def __setitem__(self, key, value):
        if self._disable_onchange:
            super().__setitem__(key, value)
            return

        to_remove = self[key]
        if not isinstance(to_remove, list):
            to_remove = (to_remove,)

        self._notify_remove(key, to_remove)
        super().__setitem__(key, value)
        if to_remove:
            self._notify_post_remove()

        added = value
        if not isinstance(added, Iterable):
            added = (added,)
        self._notify_add(key, added)

    def __repr__(self):
        return f'{type(self).__name__}({super().__repr__()})'


__all__ = ['Trackable', 'TrackableList']



from pyodide.ffi import IN_BROWSER

if not IN_BROWSER:
    import sys

    from beepy.utils import js

    sys.modules['js'] = js
    del sys, js

import beepy.children  # must be loaded before .framework due to circular import
import beepy.utils.import_hooks  # allows to use `import` for local files  # noqa: F401
from beepy.attrs import attr, html_attr, state
from beepy.components import Directive
from beepy.context import CONTENT, OVERWRITE, SUPER
from beepy.framework import Tag, __version__, empty_tag, mount
from beepy.listeners import on
from beepy.style import Style
from beepy.tags import Body, Head
from beepy.types import safe_html, safe_html_content
from beepy.utils import __CONFIG__

__all__ = [
    'Head',
    'Body',
    'Style',
    '__CONFIG__',
    'attr',
    'state',
    'html_attr',
    'on',
    'safe_html',
    'safe_html_content',
    'Directive',
    'OVERWRITE',
    'SUPER',
    'CONTENT',
    'Tag',
    'empty_tag',
    'mount',
    '__version__',
]



#!/usr/bin/env python3
import argparse
import asyncio
import functools
import http.server
import os
import re
import shutil
import socketserver
import time
from pathlib import Path
from threading import Thread

import dotenv
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from websockets.exceptions import ConnectionClosedOK
from websockets.server import serve

dotenv.load_dotenv()
BASE_DIR = Path(__file__).resolve().parent


class MonitorFolder(FileSystemEventHandler):
    def __init__(self, server):
        self.server = server

    def on_any_event(self, event):
        if not (
            event.event_type in ('opened', 'closed')
            or event.is_directory
            or event.src_path.endswith(('~', '.tmp'))
            or re.search(r'/(__pycache__|.git|.idea|dist|build)/', event.src_path)
        ):
            self.server.handle_file_event(event)


class DevServer:
    def __init__(self, *, root_path=None):
        self.websockets = []
        self.root_path = root_path
        self.observer = None
        self.developer_mode = 'DEVELOPMENT' in os.environ
        self.handle_cmd_args()

    def handle_cmd_args(self):
        parser = argparse.ArgumentParser(prog='beepy.dev', description='Simple dev server for BeePy')
        parser.add_argument(
            '-d', '--root-dir', default=Path.cwd(), help='Root directory to start server and watch file changes'
        )
        parser.add_argument('--create', action='store_true', help='Create a default .html and .py files before start')
        args = parser.parse_args()

        if not self.root_path:
            self.root_path = Path(args.root_dir)

        if not args.create:
            return

        dst = self.root_path / 'index.html'

        if dst.exists():
            print('[BeePy] Warning: File "index.html" already exists, skipping creating default files')
            return

        shutil.copyfile(BASE_DIR / 'example.html', dst)
        shutil.copyfile(BASE_DIR / 'example.py', self.root_path / '__init__.py')
        print('[BeePy] Created default files in root directory')

    async def ws_send(self, message):
        await asyncio.sleep(1)  # Hack for Django autoreload

        for ws in self.websockets[:]:
            try:
                await ws.send(message)
            except ConnectionClosedOK:
                self.websockets.remove(ws)

        if not self.websockets:
            print('[BeePy] No clients connected! Please, restart your page to connect to the dev server')

    def handle_file_event(self, event):
        path: str = event.src_path.removeprefix(str(self.root_path)).removeprefix('/')

        print(f'[BeePy] Found file change: {path}')
        if self.developer_mode:
            if path.endswith('.py'):
                os.system('hatch build')
            elif path.endswith('.js'):
                os.system(f'cd {self.root_path}/web; npm run build; cd -')  # rebuild dist
            asyncio.run(self.ws_send('__'))
        else:
            asyncio.run(self.ws_send(path))

    def _watcher_start(self):
        event_handler = MonitorFolder(self)
        observer = Observer()
        observer.schedule(event_handler, path=self.root_path, recursive=True)
        print(f'[BeePy] Monitoring started for {self.root_path}')
        observer.start()
        self.observer = observer
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
            observer.join()

    async def _ws_echo(self, websocket):
        self.websockets.append(websocket)

        async for message in websocket:
            print(f'{websocket=} {message=}')  # We really don't receive messages

    async def _ws_main(self):
        print('[BeePy] WebSockets started')
        async with serve(self._ws_echo, 'localhost', 8998):
            await asyncio.Future()  # run forever

    def _ws_start(self):
        asyncio.run(self._ws_main())

    def _simple_http_start(self, port=8888):
        # Fixes "Address already in use"
        socketserver.TCPServer.allow_reuse_address = True

        with socketserver.TCPServer(
            ('', port), functools.partial(http.server.SimpleHTTPRequestHandler, directory=self.root_path)
        ) as httpd:
            print(f'[BeePy] Serving at port {port}\nOpen server: http://localhost:{port}')
            httpd.serve_forever()

    def start(self, *, start_http=True, forever=True):
        if self.observer is not None:
            print('[BeePy] Server is already started')
            return

        Thread(target=self._ws_start, daemon=True).start()
        Thread(target=self._watcher_start, daemon=True).start()

        if not start_http:
            return

        if start_http is True:
            start_http = self._simple_http_start

        if forever:
            start_http()
        else:
            Thread(target=start_http, daemon=True).start()


if __name__ == '__main__':
    DevServer().start()



import asyncio
import inspect
from functools import wraps

import beepy
from beepy.utils.dev import _debugger
from beepy.utils.js_py import js


def ensure_sync(to_await, callback=None):
    if inspect.iscoroutine(to_await):
        task = asyncio.get_event_loop().run_until_complete(to_await)
        if callback:
            task.add_done_callback(callback)
        return task

    if callback:
        callback(to_await)
    return to_await


def ensure_sync_many(to_awaits, callback=None):
    tasks = []
    raw_results = []

    for to_await in to_awaits:
        if inspect.iscoroutine(to_await):
            tasks.append(to_await)
        else:
            raw_results.append(to_await)

    if not tasks:
        if callback:
            callback(tasks, raw_results)
        return tasks, raw_results

    task = asyncio.get_event_loop().run_until_complete(asyncio.gather(*tasks))
    if callback:
        task.add_done_callback(lambda r: callback(r, raw_results))
    return task, raw_results


def force_sync(function):
    # TODO: maybe do it with class will be better?

    callbacks = []

    @wraps(function)
    def wrapper(*args, _done_callback_=None, **kwargs):
        current_callbacks = [(cb() if dynamic else cb) for cb, dynamic in callbacks]
        current_callbacks.append(_done_callback_)

        def _callback(_res_):
            try:
                r = _res_.result()
            except Exception as e:  # noqa: BLE001 - Error in callback, just skip
                _debugger(e)
                return

            for cb in current_callbacks:
                if cb is None:
                    continue
                try:
                    cb(*args, **kwargs, _res_=r)
                except Exception as e:  # noqa: BLE001 - Error in callback, just skip
                    _debugger(e)

        return ensure_sync(function(*args, **kwargs), _callback)

    wrapper.run_after = wrapper.add_callback = lambda fn: callbacks.append((fn, False)) or fn
    wrapper.add_dynamic_callback = lambda fn: callbacks.append((fn, True)) or fn

    return wrapper


def force_sync__wait_load(function):
    wrapper = force_sync(function)
    wrapper.add_dynamic_callback(beepy.context.Context.create_onload)
    return wrapper


force_sync.wait_load = force_sync__wait_load


delay = js.delay


@force_sync
async def sleep(s):  # TODO: check if this actually works or not
    return await js.delay(s * 1000)


__all__ = ['ensure_sync', 'force_sync', 'delay', 'sleep']



from __future__ import annotations

from boltons.iterutils import first

from beepy import Tag
from beepy.attrs import attr, html_attr, state
from beepy.style import Style
from beepy.tags import div, ul
from beepy.utils import js
from beepy.utils.js_py import replace_url


class tab(div, name='tab'):
    tab_id = state(type=str)
    visible = html_attr(default=False)
    title: tab_title = state()

    default_style = Style(
        styles={
            'padding': '6px 12px',
            'animation': 'fadeEffect 1s',
            'display': 'none',
            '&[visible]': {
                'display': 'block',
            },
        }
    )

    @attr()
    def id(self) -> str:
        return f'tab-{self.parent.name}/{self.tab_id}'

    parent: tabs

    def __set_ref__(self, parent, ref):
        super().__set_ref__(parent, ref)
        self.tab_id = ref.name


class tab_title(Tag, name='li', content_tag=None):
    _tab: tab = state(type=tab)
    selected = attr(default=False)


class tabs(Tag, name='tabs'):
    dark_theme = attr(default=False)
    selected_id = attr(type=str)
    selected: tab = state(type=tab)

    name: str = 'Unknown'

    default_style = Style(
        styles={
            'a': {
                'color': 'lightskyblue',
                'text-decoration': 'none',
            },
            'ul': {
                '': '''
                list-style-type: none;
                margin: 0;
                padding: 0;
                overflow: hidden;
                background-color: #e1e1e1;
                ''',
                'li': {
                    '': '''
                    float: left;
                    font-family: "Lato", sans-serif;
                    display: inline-block;
                    text-align: center;
                    padding: 14px 16px;
                    text-decoration: none;
                    transition: 0.3s;
                    font-size: 12px;
                    ''',
                    '&:hover': {
                        'background-color': '#cdcdcd',
                    },
                    '&[selected]': {
                        'background-color': '#b1b1b1',
                    },
                },
            },
            '&[dark-theme] ul': {
                'background-color': '#1e1e1e',
                'li': {
                    '&:hover': {
                        'background-color': '#222222',
                    },
                    '&[selected]': {
                        'background-color': '#3e3e3e',
                    },
                },
            },
        }
    )

    children = [
        tabs_titles := ul(),
    ]

    @property
    def tabs_list(self) -> dict[str, tab]:
        # TODO: make tabs_list one-time calculated attribute
        return {name: _tab for name, _tab in self.ref_children.items() if isinstance(_tab, tab)}

    @classmethod
    def __class_declared__(cls):
        for tab_id, _tab_title in cls.tabs_titles.child.ref_children.items():
            _tab_title.on('click')(lambda _ul, _tab_id=tab_id: _ul.parent.select_tab(_tab_id))

    def mount(self):
        for tab_id, _tab in self.tabs_list.items():
            getattr(self.tabs_titles, tab_id).link_parent_attrs(self)

        url = js.URL.new(js.location.href)

        selected = None
        if url.hash and url.hash.startswith(f'#tab-{self.name}/'):
            selected = self.select_tab(url.hash[len(f'#tab-{self.name}/') :])
            if selected:
                js.location.hash = ''
        if not selected:
            selected = self.select_tab(url.searchParams.get(self.name))
        if not selected:
            self.select_tab(first(self.tabs_titles.ref_children.keys()))
        self._update_url()

    def select_tab(self, tab_id):
        if not tab_id or not hasattr(self, tab_id):
            return

        self.selected_id = tab_id
        if not self.selected or self.selected.tab_id != tab_id:
            for child in self.tabs_list.values():
                child.visible = False
            for title in self.tabs_titles.ref_children.values():
                title.selected = False
            self.selected = getattr(self, tab_id)
            self.selected.visible = True
            getattr(self.tabs_titles, tab_id).selected = True

        self._update_url()
        return self.selected

    def _update_url(self):
        selected_tab = self.selected

        url = js.URL.new(js.location.href)
        if (not url.hash and url.href[-1] == '#') or url.hash == '#':
            url.href = url.href[:-1]
        if url.searchParams.get(self.name) == selected_tab.tab_id:
            return

        url.searchParams.set(self.name, selected_tab.tab_id)  # modifies url.href

        replace_url(url, name=selected_tab.id, title=''.join(getattr(self.tabs_titles, selected_tab.tab_id).content()))


__all__ = ['tab', 'tab_title', 'tabs']



from __future__ import annotations

import json
from collections.abc import Mapping

from beepy.utils import js


class LocalStorage(Mapping):
    __slots__ = ('prefix',)

    def __init__(self, key: str):
        if not key.strip():
            raise TypeError('Local Storage key cannot be empty')

        self.prefix = key

    def __contains__(self, key):
        return (self.prefix + key) in js.localStorage

    def __getitem__(self, key):
        return json.loads(js.localStorage.getItem(self.prefix + key) or 'null')

    def __setitem__(self, key, value):
        js.localStorage.setItem(self.prefix + key, json.dumps(value))

    def __iter__(self):
        return (key[len(self.prefix) :] for key in js.Object.keys(js.localStorage) if key.startswith(self.prefix))

    def __len__(self):
        return len(tuple(iter(self)))


class _GlobalLocalStorage(LocalStorage):
    def __init__(self):
        super().__init__('-')
        self.prefix = ''


GlobalLocalStorage = _GlobalLocalStorage()


__all__ = ['LocalStorage', 'GlobalLocalStorage']



from beepy import Tag, mount, on, state


class IncrementButton(Tag, name='button'):
    count = state(0)

    @on
    def click(self):
        self.count += 1

    def content(self):
        return f'Count: {self.count}'


class View(Tag):
    children = [
        '🎉 Everything works!',
        IncrementButton(),
    ]


mount(View(), '#root')



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Generic, TypeVar

import beepy
from beepy.components import Component
from beepy.types import Children, ContentType, Renderer, WebBase
from beepy.utils import js, log
from beepy.utils.internal import _PY_TAG_ATTRIBUTE

if TYPE_CHECKING:
    from beepy.context import Context
    from beepy.framework import Tag
else:
    Tag = None


class CustomWrapper(WebBase):
    pass


class StringWrapper(CustomWrapper):
    __slots__ = ('content',)

    def __init__(self, string):
        self.content = string

    def __mount__(self, element, parent: Tag, index=None):
        element.insertChild(self._render(self.content).format(self=parent), index)

    def __repr__(self):
        return f'String({self.content})'


class ContentWrapper(CustomWrapper):
    __slots__ = ('content', 'tag', 'mount_element', '_current_render', 'parent', 'mount_parent', 'children')

    SHADOW_ROOTS = ('article', 'aside', 'blockquote', 'body', 'div', 'footer', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6')
    SHADOW_ROOTS += ('header', 'main', 'nav', 'p', 'section', 'span')

    content: Callable[[], ContentType | Tag]
    tag: Tag | None
    mount_element: js.HTMLElement | None
    _current_render: list[Renderer]
    parent: Tag | None
    mount_parent: js.HTMLElement | None
    children: list[Tag] | None

    def __init__(self, content, tag, _current_render):
        self.content = content
        self.tag = tag
        if tag:
            self.mount_element = tag.mount_element
        else:
            self.mount_element = None

        self._current_render = _current_render
        self.parent = None
        self.mount_parent = None
        self.children = None

    def __mount__(self, element, parent: Tag, index=None):
        self.parent = parent
        self.mount_parent = element
        if self.tag:
            self.mount_element = self.tag.clone(parent).mount_element
            self.mount_parent.insertChild(self.mount_element, index)
        else:
            self.mount_element = js.document.createDocumentFragment()
            self.mount_parent.insertChild(self.mount_element, index)
        setattr(self.mount_element, _PY_TAG_ATTRIBUTE, self)

    def _mount_children(self):
        content = self.content()

        if isinstance(content, beepy.framework.Tag):
            content = (content,)
        elif isinstance(content, Iterable) and not isinstance(content, str) and content:
            content = tuple(content)
            for _child in content[:]:
                if not isinstance(_child, beepy.framework.Tag):
                    content = None
                    break
        else:
            content = None

        if content:
            for child in content:
                child.__mount__(self.mount_element, self.parent)
            self.children = content

    def __unmount__(self, element, parent):
        if self.mount_parent is not element:
            return

        if self.tag:
            element.safeRemoveChild(self.mount_element)
        # TODO: handle shadowRoot and documentFragment?

    def __render__(self):
        for renderer in self._current_render:
            if self.parent and renderer not in self.parent._dependents:
                self.parent._dependents.append(renderer)

        self._current_render.append(self)

        if self.children:
            for child in self.children:
                child.__render__()

            if self._current_render[-1] is self:
                self._current_render.pop()
            return

        result = self._render(self.content())
        if not isinstance(result, str):
            raise TypeError(f'Function {self.content} cannot return {result}!')

        if self.tag or (self.parent and self.parent._shadow_root):
            self.mount_element.innerHTML = result
        else:  # fragment can't be re-rendered
            self.mount_element.innerHTML = result
            current_html = self.mount_parent.innerHTML
            current_html_escaped = self._render(current_html)
            if result and result not in (current_html, current_html_escaped):
                if current_html and not (self.parent and self.parent._raw_html):
                    log.warn(
                        f'This html `{current_html}` will be replaces with this: `{result}`.\n'
                        'Maybe you must use beepy.Tag instead of beepy.tags.div, '
                        'or you used incorrect html tags like <br/> instead of <br>',
                    )
                self.mount_parent.innerHTML = result

        if self._current_render[-1] is self:
            self._current_render.pop()

    def __repr__(self):
        return f'<{self.parent}.{self.content.__name__}()>'


C = TypeVar('C')


class ChildRef(WebBase, Generic[C]):
    __slots__ = ('name', 'child', 'inline_def', '_cache')

    name: str | None
    child: C
    _cache: dict[Context, C]

    def __init__(self, child: C, *, inline_def=False):
        self.name = None
        self.child = child
        self.inline_def = inline_def
        self._cache = {}

    def __repr__(self):
        return f'{type(self).__name__}(Tag.{self.name} = {self.child})'

    def __get__(self, instance: Context | None, owner: type[Context] | None = None) -> ChildRef | C:
        if instance is None:
            return self

        if (result := self._cache.get(instance)) is not None:
            return result

        self._cache[instance] = self.child
        return self.child

    def __set__(self, instance: Context, value: C):
        self._cache[instance] = value

    def __delete__(self, instance: Context):
        del self._cache[instance]

    def __set_name__(self, owner: type[Tag], name: str):
        self.name = name

    @abstractmethod
    def _update_child(self, parent: Tag, index: int):
        pass

    def __render__(self, parent: Tag):
        self.__get__(parent).__render__()

    def __mount__(self, element, parent: Tag, index=None):
        self.__get__(parent).__mount__(element, parent, index)

    def __unmount__(self, element, parent: Tag):
        self.__get__(parent).__unmount__(element, parent)


class ComponentRef(ChildRef[Component]):
    __slots__ = ()

    child: Component

    def _update_child(self, parent: Tag, index):  # noqa: ARG002 - arguments for overriding
        clone = self.__get__(parent).clone(parent)
        clone.__set_ref__(parent, self)
        self.__set__(parent, clone)
        return clone


class TagRef(ComponentRef, ChildRef[Tag]):
    __slots__ = ()

    child: Tag


class ChildrenRef(ChildRef):
    __slots__ = ()

    child: Children

    def _update_child(self, parent, index):
        copy = self.__get__(parent).copy()
        copy.__set_parent__(parent, index, self)
        self.__set__(parent, copy)
        return copy

    def __set__(self, instance, value):
        if isinstance(value, Children):
            super().__set__(instance, value)
        else:
            current_value = self.__get__(instance)
            current_value[:] = value


from beepy.framework import Tag  # noqa: E402, isort: skip - circular import

__all__ = ['CustomWrapper', 'StringWrapper', 'ContentWrapper', 'TagRef', 'ChildrenRef']



from __future__ import annotations

import inspect
from collections import defaultdict
from functools import partial, wraps
from types import MethodType
from typing import TYPE_CHECKING, Any

from beepy.utils import js, to_js
from beepy.utils.common import nested_copy
from beepy.utils.internal import _PY_TAG_ATTRIBUTE
from beepy.utils.js_py import create_proxy

if TYPE_CHECKING:
    from collections.abc import Callable

    from beepy.components import Component

global_events = {
    'keyup': js.document,
    'keypress': js.document,
    'keydown': js.document,
    'popstate': js.window,
    'hashchange': js.window,
}
_key_codes = {
    'esc': (27,),
    'tab': (9,),
    'enter': (13,),
    'space': (32,),
    'up': (38,),
    'left': (37,),
    'right': (39,),
    'down': (40,),
    'delete': (8, 46),
}


# TODO: click.(right|middle) ; click.left ; etc.


def key_code_check(key_name, event):
    return event.keyCode in _key_codes[key_name]


RAW_JS_CHECKS = ('prevent', 'stop', 'stop_all')


class on:
    __slots__ = ('name', 'callback', 'pass_event', 'child_restrict', 'modifiers', 'checks', 'js_checks')

    name: str | None
    callback: Callable[[Component, ...], Any]
    pass_event: bool
    child_restrict: type[Component] | None
    modifiers: list[str]
    checks: list[Callable[[js.Event], bool]]
    js_checks: list[str]

    def __init__(self, method):  # TODO: add 'mount' callbacks?
        self.child_restrict = None
        self.pass_event = True
        self.modifiers = []
        self.checks = []
        self.js_checks = []

        if isinstance(method, str):
            if '.' in method:
                method, *self.modifiers = method.split('.')
                for modifier in self.modifiers:
                    if modifier in _key_codes:
                        # TODO: check for visibility?
                        self.checks.append(partial(key_code_check, modifier))
                    elif modifier in RAW_JS_CHECKS:
                        self.js_checks.append(modifier)
                    else:
                        raise ValueError(f'Unknown event modifier ".{modifier}"!')

            self.name = method
            return

        self.name = None
        self(method)

    def __call__(self, method, child=None):
        self.callback = method
        self.child_restrict = child

        sig = inspect.signature(method)
        self.pass_event = 'event' in sig.parameters
        return self

    def _get_cb_and_instance(self, cmpt):
        if (
            self.child_restrict
            and cmpt.parent_defined
            and (self.child_restrict == cmpt or (cmpt._ref and self.child_restrict == cmpt._ref.child))
        ):
            cmpt = cmpt.parent

        if isinstance(self.callback, MethodType):
            return self.callback, cmpt
        else:
            return MethodType(self.callback, cmpt), cmpt

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return self._get_cb_and_instance(instance)[0]

    def _prepare_call(self, cmpt, event):
        for check in self.checks:
            if not check(event):
                return

        return self._get_cb_and_instance(cmpt)

    async def _a_call(self, cmpt, event):
        if (prepare := self._prepare_call(cmpt, event)) is None:
            return

        fn, cmpt = prepare

        args = (event,) if self.pass_event else ()
        data = await fn(*args)
        self._after_call(cmpt, event)
        return data

    def _call(self, cmpt, event):
        if (prepare := self._prepare_call(cmpt, event)) is None:
            return

        fn, cmpt = prepare
        args = (event,) if self.pass_event else ()
        data = fn(*args)
        self._after_call(cmpt, event)
        return data

    def _after_call(self, cmpt, event):
        for dependent in cmpt._dependents:
            # TODO: move to other place
            dependent.__render__()

        getattr(event.currentTarget, _PY_TAG_ATTRIBUTE, cmpt).__render__()

    def _make_listener(self, event_name: str, cmpt: Component):
        if inspect.iscoroutinefunction(self.callback):

            @wraps(self.callback)
            async def method(event):
                return await self._a_call(cmpt, event)

        else:

            @wraps(self.callback)
            def method(event):
                return self._call(cmpt, event)

        return js.beepy.addAsyncListener(
            global_events.get(event_name, cmpt.mount_element), event_name, create_proxy(method), to_js(self.js_checks)
        )

    @classmethod
    def _remove_listener(cls, event_name: str, cmpt: Component, event_listener: Callable):
        global_events.get(event_name, cmpt.mount_element).removeEventListener(event_name, event_listener)

    def __set__(self, instance, value):
        raise AttributeError(f'Cannot set on_{self.name} event handler')

    def __delete__(self, instance):
        pass

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name

        if self.child_restrict is None:
            owner._static_listeners = defaultdict(list, **nested_copy(owner._static_listeners))
            owner._static_listeners[self.name].append(self)

    def __repr__(self):
        return f'on_{self.name}({self.callback})'


__all__ = ['on', '_key_codes', 'RAW_JS_CHECKS']



from __future__ import annotations

import re
from dataclasses import dataclass
from re import Match

from beepy import Tag
from beepy.attrs import html_attr, state
from beepy.listeners import on
from beepy.tags import a
from beepy.types import Children
from beepy.utils import js
from beepy.utils.dev import _debugger
from beepy.utils.internal import lazy_import_cls, reload_requirements
from beepy.utils.js_py import push_url


class WithRouter:
    match: Match = state(move_on=True)
    router: Router = state(move_on=True)


@dataclass
class Path:
    # TODO: move to utils?
    pathname: str = ''
    search: str = ''
    hash: str = ''

    @classmethod
    def parse(cls, path: str):
        result = cls()

        if path:
            if '#' in path:
                path, hash = path.split('#', 1)
                result.hash = '#' + hash

            if '?' in path:
                path, search = path.split('?', 1)
                result.search = '?' + search

            if path:
                result.pathname = path

        return result

    @classmethod
    def parse_to(cls, to):
        if not to:
            return '/'
        if isinstance(to, str):
            to = cls.parse(to)
        return to.pathname

    def iter_search(self):
        for search_element in self.search[1:].split('&'):
            if not search_element:
                continue

            key, value = search_element.split('=', 1)
            yield key, js.decodeURIComponent(value)

    def push_state(self):
        url = js.URL.new(js.location.origin + self.pathname + self.hash)

        if (not url.hash and url.href[-1] == '#') or url.hash == '#':
            url.href = url.href[:-1]

        for key, value in self.iter_search():
            url.searchParams.set(key, value)  # modifies url.href

        push_url(url, path=self.__dict__)


class Link(a, WithRouter):
    to = state(type=str, required=True)
    href = html_attr(type=str)

    @to.on('init', 'change')
    def to_changed(self, value):
        if self.router:
            value = self.router.basename + value
        self.href = Path.parse_to(value)

    @on('click.prevent')
    async def navigate(self):
        Path.parse(self.router.basename + self.to).push_state()
        await self.router._history_refresh()


class Router(Tag):
    basename = ''
    routes: dict[str, str | type[Tag]] = {
        # r'/$': Tag,
        # r'/app/(?P<id>.*)$': 'app.App',  # lazy import!
    }

    fallback_tag_cls = None
    single_tag = True

    children = [
        components := Children(),
    ]

    def pre_mount(self):
        self._load_children()

    @on('popstate')
    async def _history_refresh(self):
        js.beepy.startLoading(mountPoint=self._root_parent.mount_element)
        await reload_requirements()
        self._load_children()
        js.beepy.stopLoading()

    def import_tag_component(self, tag_cls: str | type[Tag], match, **kwargs):
        try:
            tag_cls: type[Tag] = lazy_import_cls(tag_cls)
        except ModuleNotFoundError as e:
            _debugger(e)
            raise

        if issubclass(tag_cls, WithRouter):
            kwargs |= {'router': self, 'match': match}
        return tag_cls(**kwargs)

    def add_tag_component(self, tag_cls: str | type[Tag], match, path):  # noqa: ARG002 - arguments for overriding
        self.components.append(self.import_tag_component(tag_cls, match=match))

    def _load_children(self):
        self._current_render.clear()

        old_components = list(self.components)

        with self.components._disable_onchange:  # can Locker also be descriptor with auto-replace as in last two lines?
            self.components.clear()

            for path, tag_cls in self.routes.items():
                if match := re.search(self.basename + path, js.location.pathname):
                    self.add_tag_component(tag_cls, match=match, path=path)
                    if self.single_tag:
                        break

            if not self.components:
                if fallback := self.fallback_tag_cls:
                    self.add_tag_component(fallback, match=None, path=None)
                else:
                    # TODO: maybe create BeePyError?
                    raise ValueError('No route to use!')

            for child in self.components:
                child.link_parent_attrs(self)
                args, kwargs = child.args_kwargs
                kwargs = child._attrs_defaults | kwargs
                child.init(*args, _load_children=False, **kwargs)

            new_components, self.components = list(self.components), old_components
        self.components[:] = new_components  # triggers correct onchange handlers
